# PYDANTIC
from pydantic import Field
from pydantic_settings import BaseSettings


def create_internal_config(
    token_url_alias: str = "ALPHA_TOKEN_URL",
    user_me_url_alias: str = "ALPHA_USER_ME_URL",
    api_key_me_url_alias: str = "ALPHA_API_KEY_ME_URL",
    secret_key_alias: str = "ALPHA_SECRET_KEY",
    algorithm_alias: str = "ALPHA_ALGORITHM",
):
    class AlphaInternalConfigSettingsSchema(BaseSettings):
        token_url: str = Field(validation_alias=token_url_alias)
        user_me_url: str = Field(validation_alias=user_me_url_alias)
        api_key_me_url: str = Field(validation_alias=api_key_me_url_alias)
        secret_key: str = Field(validation_alias=secret_key_alias)
        algorithm: str = Field(validation_alias=algorithm_alias)

    return AlphaInternalConfigSettingsSchema()



# MODULES
from typing import Optional

# PYDANTIC
from pydantic import BaseModel, ConfigDict, Field


class ContactSchema(BaseModel):
    name: Optional[str] = Field(default=None)
    email: Optional[str] = Field(default=None)


class OpenApiSchema(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    description: Optional[str] = Field(default=None)
    contact: Optional[ContactSchema] = Field(default=None)



from alphaz_next.core.responses.file_response import FileResponse
from alphaz_next.core.responses.html_response import HTMLResponse
from alphaz_next.core.responses.json_response import JSONResponse
from alphaz_next.core.responses.orjson_response import ORJSONResponse
from alphaz_next.core.responses.plaintext_response import PlainTextResponse
from alphaz_next.core.responses.redirect_response import RedirectResponse
from alphaz_next.core.responses.streaming_response import StreamingResponse
from alphaz_next.core.responses.ujson_response import UJSONResponse
from alphaz_next.core.responses.response import Response



# PYDANTIC
from typing import Annotated

# PYDANTIC
from pydantic import (
    BaseModel,
    ConfigDict,
    FieldValidationInfo,
    PositiveInt,
    StringConstraints,
    computed_field,
    field_validator,
)

LOGGING_LEVEL = {
    "CRITICAL": 50,
    "FATAL": 50,
    "ERROR": 40,
    "WARNING": 30,
    "WARN": 30,
    "INFO": 20,
    "DEBUG": 10,
    "NOTSET": 0,
}


class TimeRotatingSchema(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    when: str
    interval: PositiveInt
    backup_count: PositiveInt


class LoggingSchema(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    level: Annotated[
        str,
        StringConstraints(
            strip_whitespace=True,
            to_upper=True,
        ),
    ]
    time_rotating: TimeRotatingSchema

    @field_validator("level")
    @classmethod
    def validate_level(cls, value: str, info: FieldValidationInfo):
        if value not in LOGGING_LEVEL:
            raise ValueError(f"{info.field_name} is not valid")

        return value

    @computed_field
    @property
    def level_code(self) -> int:
        return LOGGING_LEVEL.get(self.level)



# MODULES
from typing import Annotated, Any, Dict, List

# FASTAPI
from fastapi import Depends, status
from fastapi.security import APIKeyHeader, OAuth2PasswordBearer, SecurityScopes

# JOSE
from jose import JWTError, jwt

# LIBS
from alphaz_next.libs.httpx import (
    make_async_request_with_retry,
    post_process_http_response,
)

# MODELS
from alphaz_next.models.auth.user import UserSchema, UserShortSchema
from alphaz_next.models.config._base.internal_config_settings import (
    create_internal_config,
)

# EXCEPTIONS
from alphaz_next.core.exceptions import (
    InvalidCredentialsError,
    NotEnoughPermissionsError,
    HTTPException,
)

INTERNAL_CONFIG = create_internal_config()

API_KEY_HEADER = APIKeyHeader(name="api_key", auto_error=False)
OAUTH2_SCHEME = OAuth2PasswordBearer(tokenUrl=INTERNAL_CONFIG.token_url)


def decode_token(token: str) -> Dict[str, Any]:
    try:
        payload = jwt.decode(
            token,
            INTERNAL_CONFIG.secret_key,
            algorithms=[INTERNAL_CONFIG.algorithm],
        )
    except JWTError:
        raise InvalidCredentialsError()

    username: str = payload.get("sub")
    if username is None:
        raise InvalidCredentialsError()

    return payload


async def get_user(token: str) -> UserSchema:
    decode_token(token=token)

    headers = {
        "Authorization": f"Bearer {token}",
    }

    response = await make_async_request_with_retry(
        method="POST",
        url=INTERNAL_CONFIG.user_me_url,
        **{
            "headers": headers,
        },
    )

    return post_process_http_response(
        response,
        schema=UserSchema,
    )


async def get_api_key(api_key: str) -> UserShortSchema:
    headers = {
        "api_key": api_key,
    }

    response = await make_async_request_with_retry(
        method="POST",
        url=INTERNAL_CONFIG.api_key_me_url,
        **{
            "headers": headers,
        },
    )

    return post_process_http_response(
        response,
        schema=UserShortSchema,
    )


def check_user_permissions(permissions: List[str], user_permissions: List[str]) -> None:
    if len(permissions) > 0 and not any(
        [user_permission in permissions for user_permission in user_permissions]
    ):
        raise NotEnoughPermissionsError()


async def get_user_from_jwt(
    security_scopes: SecurityScopes,
    token: Annotated[str, Depends(OAUTH2_SCHEME)],
) -> UserSchema:
    try:
        user = await get_user(token=token)

        check_user_permissions(
            permissions=security_scopes.scopes,
            user_permissions=user.permissions,
        )

        return user

    except InvalidCredentialsError as ex:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            headers={
                "WWW-Authenticate": "Bearer",
            },
            ext_headers={
                "status_description": ex.args,
            },
        )
    except NotEnoughPermissionsError as ex:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            headers={
                "WWW-Authenticate": "Bearer",
            },
            ext_headers={
                "status_description": ex.args,
            },
        )


async def get_user_from_api_key(
    security_scopes: SecurityScopes,
    api_key: Annotated[
        str,
        Depends(API_KEY_HEADER),
    ],
) -> UserShortSchema:
    try:
        if api_key is None:
            raise InvalidCredentialsError()

        user = await get_api_key(api_key=api_key)

        check_user_permissions(
            permissions=security_scopes.scopes,
            user_permissions=user.permissions,
        )

        return user

    except InvalidCredentialsError as ex:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            headers={
                "WWW-Authenticate": "Bearer",
            },
            ext_headers={
                "status_description": ex.args,
            },
        )
    except NotEnoughPermissionsError as ex:
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            headers={
                "WWW-Authenticate": "Bearer",
            },
            ext_headers={
                "status_description": ex.args,
            },
        )



# MODULES
from os import PathLike, stat_result
from typing import Mapping

# FASTAPI
from fastapi.responses import FileResponse as _FileResponse

# STARLETTE
from starlette.background import BackgroundTask

# CORE
from alphaz_next.core._base import extend_headers, ExtHeaders


class FileResponse(_FileResponse):
    def __init__(
        self,
        path: str | PathLike[str],
        status_code: int = 200,
        headers: Mapping[str, str] | None = None,
        ext_headers: ExtHeaders | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
        filename: str | None = None,
        stat_result: stat_result | None = None,
        method: str | None = None,
        content_disposition_type: str = "attachment",
    ) -> None:
        headers = extend_headers(
            headers=headers,
            ext_headers=ext_headers,
        )

        super().__init__(
            path,
            status_code,
            headers,
            media_type,
            background,
            filename,
            stat_result,
            method,
            content_disposition_type,
        )



# MODULES
import getpass
import os
from typing import Any, Dict, TypedDict
from pathlib import Path

# PYDANTIC
from pydantic import BaseModel, ConfigDict, model_validator

# MODELS
from alphaz_next.models.config.api_config import AlphaApiConfigSchema


class ReservedConfigItem(TypedDict):
    root: str
    project_name: str


class AlphaConfigSchema(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    environment: str
    config_file_path: Path
    project_name: str
    version: str
    root: str

    api_config: AlphaApiConfigSchema

    @model_validator(mode="before")
    @classmethod
    def validate_model(cls, data: Dict[str, Any]) -> Dict:
        tmp = replace_reserved_config(
            data,
            reserved_config=ReservedConfigItem(
                root=data.get("root"),
                project_name=data.get("project_name"),
            ),
        )

        reserved_fields = ReservedConfigItem(
            root=tmp.get("root"),
            project_name=tmp.get("project_name"),
        )

        for key, value in tmp.items():
            if isinstance(value, dict):
                tmp[key]["__reserved_fields__"] = reserved_fields

        return tmp


def replace_reserved_config(
    config: Dict,
    reserved_config: ReservedConfigItem,
) -> Dict:
    replaced_config = config.copy()

    def replace_variable(value: Any):
        return (
            (
                value.replace("{{root}}", reserved_config.get("root"))
                .replace("{{home}}", os.path.expanduser("~"))
                .replace("{{project_name}}", reserved_config.get("project_name"))
                .replace("{{user}}", getpass.getuser())
                .replace("{{project}}", os.path.abspath(os.getcwd()))
            )
            if isinstance(value, str)
            else value
        )

    def traverse(obj):
        if isinstance(obj, dict):
            for key, value in obj.items():
                if isinstance(value, (dict, list)):
                    traverse(value)
                else:
                    obj[key] = replace_variable(value)
        elif isinstance(obj, list):
            for i, value in enumerate(obj):
                if isinstance(value, (dict, list)):
                    traverse(value)
                else:
                    obj[i] = replace_variable(value)

        return obj

    return traverse(replaced_config)



# MODULES
from typing import Optional

# PYDANTIC
from pydantic import BaseModel, ConfigDict, Field

# MODELS
from alphaz_next.models.config.apm_config import ApmConfig
from alphaz_next.models.config.logging_config import LoggingSchema
from alphaz_next.models.config.openapi_config_schema import OpenApiSchema


class AlphaApiConfigSchema(BaseModel):
    model_config = ConfigDict(
        from_attributes=True,
        extra="allow",
    )

    databases_config_path: str
    port: int
    workers: int
    logging: LoggingSchema
    apm: Optional[ApmConfig] = Field(default=None)
    openapi: Optional[OpenApiSchema] = Field(default=None)



# MODULES
from pathlib import Path
from typing import Type, TypeVar

# PYDANTIC
from pydantic import Field, computed_field
from pydantic_settings import BaseSettings

# LIBS
from alphaz_next.libs.file_lib import open_json_file
from alphaz_next.models.config.alpha_config import AlphaConfigSchema

_T = TypeVar("_T", bound=AlphaConfigSchema)


def create_config_settings(
    model: Type[_T],
    environment_alias: str = "ALPHA_ENV",
    config_dir_alias: str = "ALPHA_CONFIG_DIR",
):
    class AlphaConfigSettingsSchema(BaseSettings):
        environment: str = Field(validation_alias=environment_alias)
        config_dir: str = Field(validation_alias=config_dir_alias)

        @computed_field
        @property
        def main_config(self) -> _T:
            config_file_path = Path(self.config_dir) / f"config.{self.environment}.json"

            data = open_json_file(path=config_file_path)

            data_ext = {
                "environment": self.environment,
                "config_file_path": config_file_path,
            }

            data.update(data_ext)

            return model.model_validate(data)

    return AlphaConfigSettingsSchema().main_config



# MODULES
import json
from pathlib import Path
from typing import Any


def open_json_file(path: Path, encoding="utf-8"):
    if not path.exists():
        raise FileNotFoundError(f"Path {path} does not exist")

    if not path.is_file():
        raise FileExistsError(f"Path {path} is not a file")

    with open(path, encoding=encoding) as json_file:
        raw_data: dict = json.load(json_file)

    return raw_data


def save_json_file(
    path: Path,
    data: Any,
    encoding: str = "utf-8",
):
    if path.exists():
        return

    with open(path, "w", encoding=encoding) as file:
        file.write(json.dumps(data))


def save_file(
    path: Path,
    data: bytes,
    encoding: str = "utf-8",
    new_line: str = "\n",
):
    if path.exists():
        return

    text_data = data.decode(encoding)

    with open(path, "w", encoding=encoding, newline=new_line) as file:
        file.write(text_data)



# MODULES
import asyncio
import time
from typing import Any, List, Literal, Optional, Type, TypeVar, Union

# FASTAPI
from fastapi import HTTPException

# HTTPX
import httpx
from pydantic import BaseModel


def make_request_with_retry(
    method: Literal["POST", "PATCH", "PUT", "DELETE", "GET"],
    url,
    max_retries: int = 3,
    retry_on_status: Optional[List[int]] = None,
    timeout: Optional[float] = None,
    **kwargs,
) -> httpx.Response:
    """
    Makes an HTTP request with retries in case of a timeout error.

    Args:
        method (Literal["POST", "PATCH", "PUT", "DELETE", "GET"]): The HTTP method to use.
        url (str): The URL to make the request to.
        max_retries (int, optional): The maximum number of retries to attempt. Defaults to 3.
        retry_on_status (List[int], optional): A list of HTTP status codes to retry on. Defaults to None.
        timeout (float, optional): The timeout value for the request. Defaults to None.
        **kwargs: Additional keyword arguments to pass to the HTTP client.

    Returns:
        httpx.Response: The HTTP response.

    Raises:
        RuntimeError: If the maximum number of retries is exceeded or unknown error occurs.
    """
    item_repr = {
        "method": method,
        "url": url,
        "kwargs": kwargs,
    }

    retry_statuses = retry_on_status or []

    for retry_count in range(max_retries + 1):
        try:
            with httpx.Client(timeout=timeout) as client:
                response: httpx.Response = getattr(client, method.lower())(
                    url, **kwargs
                )
        except httpx.ReadTimeout as ex:
            if retry_count == max_retries:
                raise RuntimeError(
                    f"Unable to contact server after {retry_count} retries {item_repr}"
                )

            wait_time = 2**retry_count
            time.sleep(wait_time)
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code in retry_statuses and retry_count < max_retries:
                wait_time = 2**retry_count
                time.sleep(wait_time)
            else:
                raise ex
        except Exception as ex:
            raise RuntimeError(
                f"An unknown error occurs while contacting external server {item_repr}"
            )
        else:
            return response

    raise RuntimeError(f"Maximum number of retries exceeded {item_repr}")


async def make_async_request_with_retry(
    method: Literal["POST", "PATCH", "PUT", "DELETE", "GET"],
    url: str,
    max_retries: int = 3,
    retry_on_status: Optional[List[int]] = None,
    timeout: Optional[float] = None,
    **kwargs,
) -> httpx.Response:
    """
    Makes an HTTP request with retries in case of a timeout error.

    Args:
        method (Literal["POST", "PATCH", "PUT", "DELETE", "GET"]): The HTTP method to use.
        url (str): The URL to make the request to.
        max_retries (int, optional): The maximum number of retries to attempt. Defaults to 3.
        retry_on_status (List[int], optional): A list of HTTP status codes to retry on. Defaults to None.
        timeout (float, optional): The timeout value for the request. Defaults to None.
        **kwargs: Additional keyword arguments to pass to the HTTP client.

    Returns:
        httpx.Response: The HTTP response.

    Raises:
        RuntimeError: If the maximum number of retries is exceeded or unknown error occurs.
    """
    item_repr = {
        "method": method,
        "url": url,
        "kwargs": kwargs,
    }

    retry_statuses = retry_on_status or []

    for retry_count in range(max_retries + 1):
        try:
            async with httpx.AsyncClient(timeout=timeout) as client:
                response: httpx.Response = await getattr(client, method.lower())(
                    url, **kwargs
                )
        except httpx.ReadTimeout as ex:
            if retry_count == max_retries:
                raise RuntimeError(
                    f"Unable to contact server after {retry_count} retries {item_repr}"
                )

            wait_time = 2**retry_count
            await asyncio.sleep(wait_time)
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code in retry_statuses and retry_count < max_retries:
                wait_time = 2**retry_count
                await asyncio.sleep(wait_time)
            else:
                raise ex
        except Exception as ex:
            raise RuntimeError(
                f"An unknown error occurs while contacting external server {item_repr}"
            )
        else:
            return response

    raise RuntimeError(f"Maximum number of retries exceeded {item_repr}")


T = TypeVar("T", bound=BaseModel)


def post_process_http_response(
    response: httpx.Response,
    schema: Type[T] = None,
    mode_alpha: bool = False,
) -> Union[T, List[T], Any]:
    """
    Processes an HTTP response and returns the response body as a validated object or a list of validated objects.

    Args:
        response: The HTTP response to process.
        schema: The schema to use for validating the response body.
        mode_alpha: Whether to use the "data" key in the response body for validation.

    Returns:
        The validated response body as an object or a list of objects.

    Raises:
        HTTPException: If the response is not a success.
    """
    if not response.is_success:
        raise HTTPException(
            status_code=response.status_code,
            detail=response.text,
            headers=response.headers,
        )

    response_json = response.json()
    if mode_alpha:
        response_json = response_json.get("data")

    if schema is None:
        return response_json

    if isinstance(response_json, (list, tuple)):
        return [schema.model_validate(item) for item in response_json]

    return schema.model_validate(response_json)



# MODULES
from typing import Any, Dict

# FASTAPI
from fastapi.exceptions import HTTPException as _HTTPException

# CORE
from alphaz_next.core._base import extend_headers, ExtHeaders


class InvalidCredentialsError(Exception):
    def __init__(self):
        super().__init__("Could not validate credentials")


class NotEnoughPermissionsError(Exception):
    def __init__(self):
        super().__init__("Not enough permissions")


class HTTPException(_HTTPException):
    def __init__(
        self,
        status_code: int,
        detail: Any = None,
        headers: Dict[str, str] | None = None,
        ext_headers: ExtHeaders | None = None,
    ) -> None:
        headers = extend_headers(
            headers=headers,
            ext_headers=ext_headers,
        )

        super().__init__(status_code, detail, headers)



# MODULES
from typing import Optional

# PYDANTIC
from pydantic import BaseModel, ConfigDict, Field, model_validator


class ApmConfig(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    server_url: Optional[str] = Field(default=None)
    environment: Optional[str] = Field(default=None)
    ssl_ca_cert: Optional[str] = Field(default=None)
    ssl_verify: bool = Field(default=True)
    debug: bool = Field(default=True)
    active: bool = Field(default=False)

    @model_validator(mode="after")
    def validate_model(self):
        if not self.active:
            return self

        if self.server_url is None:
            raise ValueError(f"server_url cannot be None if {self.active=}")

        if self.environment is None:
            raise ValueError(f"environment cannot be None if {self.active=}")

        return self



# MODULES
from enum import Enum
import json
import re
from pathlib import Path
from typing import Any, Dict, List, Optional

# DEEPDIFF
from deepdiff import DeepDiff

# UNITTEST
from unittest import TestCase

# FASTAPI
from fastapi.testclient import TestClient

# PYDANTIC
from pydantic import BaseModel, Field

# HTTPX
from httpx import Response
from httpx._types import HeaderTypes, QueryParamTypes

# LIBS
from alphaz_next.libs.file_lib import save_file, save_json_file


class ExpectedResponse(BaseModel):
    status_code: int
    data: Optional[Any] = Field(default=None)
    status_description: Optional[List[str]] = Field(default_factory=lambda: [])
    pagination: Optional[str] = Field(default=None)
    warning: Optional[bool] = Field(default=None)


class APiResponse(BaseModel):
    status_code: int
    data: Optional[Any]
    headers: Dict


class ResponseFormatEnum(Enum):
    JSON = "json"
    BYTES = "bytes"
    NO_CONTENT = "no_content"


class AlphaTestCase(TestCase):
    __RESET_BEFORE_NEXT_TEST__: bool = False

    def setUp(self):
        if not AlphaTestCase.__RESET_BEFORE_NEXT_TEST__:
            return

        self.reset_tables()

        AlphaTestCase.__RESET_BEFORE_NEXT_TEST__ = False

    @classmethod
    def setUpClass(cls) -> None:
        cls.app = cls.create_app()

        cls.client = TestClient(cls.app)

    def mark_reset_before_next_test(self):
        AlphaTestCase.__RESET_BEFORE_NEXT_TEST__ = True

    @classmethod
    def reset_tables(cls):
        raise NotImplementedError()

    @classmethod
    def create_app(cls):
        raise NotImplementedError()

    @classmethod
    def get_ignored_keys(cls) -> List[str]:
        return []

    def add_fake_headers(
        self,
        headers: dict,
        with_fake_token: bool = False,
        with_fake_api_key: bool = False,
    ):
        if headers is None:
            headers = {}

        if with_fake_token:
            headers.update(
                {"Authorization": "Bearer fake_jwt"},
            )

        if with_fake_api_key:
            headers.update(
                {"api_key": "fake_api_key"},
            )

        return headers

    def get_client(
        self,
        url: str,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[HeaderTypes] = None,
        response_format: ResponseFormatEnum = ResponseFormatEnum.JSON,
        with_fake_token: bool = True,
        with_fake_api_key: bool = True,
        saved_path: Optional[Path] = None,
    ):
        headers = self.add_fake_headers(
            headers=headers,
            with_fake_token=with_fake_token,
            with_fake_api_key=with_fake_api_key,
        )

        response = self.client.get(
            url,
            params=params,
            headers=headers,
        )

        status_code, data = self._post_process_response(
            response=response,
            response_format=response_format,
            saved_path=saved_path,
        )

        return APiResponse(
            status_code=status_code,
            data=data,
            headers=response.headers,
            header_status_description=response.headers.get("x-status-description"),
            header_pagination=response.headers.get("x-pagination"),
            header_warning=response.headers.get("x-warning"),
        )

    def put_client(
        self,
        url: str,
        data=None,
        json: Any = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[HeaderTypes] = None,
        response_format: ResponseFormatEnum = ResponseFormatEnum.JSON,
        with_fake_token: bool = True,
        with_fake_api_key: bool = True,
        saved_path: Path = None,
    ):
        headers = self.add_fake_headers(
            headers=headers,
            with_fake_token=with_fake_token,
            with_fake_api_key=with_fake_api_key,
        )

        response = self.client.put(
            url,
            data=data,
            json=json,
            params=params,
            headers=headers,
        )

        status_code, data = self._post_process_response(
            response=response,
            response_format=response_format,
            saved_path=saved_path,
        )

        return APiResponse(
            status_code=status_code,
            data=data,
            headers=response.headers,
            header_status_description=response.headers.get("x-status-description"),
            header_pagination=response.headers.get("x-pagination"),
            header_warning=response.headers.get("x-warning"),
        )

    def patch_client(
        self,
        url: str,
        data=None,
        json: Any = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[HeaderTypes] = None,
        response_format: ResponseFormatEnum = ResponseFormatEnum.JSON,
        with_fake_token: bool = True,
        with_fake_api_key: bool = True,
        saved_path: Path = None,
    ):
        headers = self.add_fake_headers(
            headers=headers,
            with_fake_token=with_fake_token,
            with_fake_api_key=with_fake_api_key,
        )

        response = self.client.patch(
            url,
            data=data,
            json=json,
            params=params,
            headers=headers,
        )

        status_code, data = self._post_process_response(
            response=response,
            response_format=response_format,
            saved_path=saved_path,
        )

        return APiResponse(
            status_code=status_code,
            data=data,
            headers=response.headers,
            header_status_description=response.headers.get("x-status-description"),
            header_pagination=response.headers.get("x-pagination"),
            header_warning=response.headers.get("x-warning"),
        )

    def post_client(
        self,
        url: str,
        data=None,
        json: Any = None,
        params: Optional[QueryParamTypes] = None,
        headers: Optional[HeaderTypes] = None,
        response_format: ResponseFormatEnum = ResponseFormatEnum.JSON,
        with_fake_token: bool = True,
        with_fake_api_key: bool = True,
        saved_path: Path = None,
    ):
        headers = self.add_fake_headers(
            headers=headers,
            with_fake_token=with_fake_token,
            with_fake_api_key=with_fake_api_key,
        )

        response = self.client.post(
            url,
            data=data,
            json=json,
            params=params,
            headers=headers,
        )

        status_code, data = self._post_process_response(
            response=response,
            response_format=response_format,
            saved_path=saved_path,
        )

        return APiResponse(
            status_code=status_code,
            data=data,
            headers=response.headers,
            header_status_description=response.headers.get("x-status-description"),
            header_pagination=response.headers.get("x-pagination"),
            header_warning=response.headers.get("x-warning"),
        )

    def delete_client(
        self,
        url: str,
        params: dict = None,
        headers: Optional[HeaderTypes] = None,
        response_format: ResponseFormatEnum = ResponseFormatEnum.JSON,
        with_fake_token: bool = True,
        with_fake_api_key: bool = True,
        saved_path: Path = None,
    ):
        headers = self.add_fake_headers(
            headers=headers,
            with_fake_token=with_fake_token,
            with_fake_api_key=with_fake_api_key,
        )

        response = self.client.delete(
            url,
            params=params,
            headers=headers,
        )

        status_code, data = self._post_process_response(
            response=response,
            response_format=response_format,
            saved_path=saved_path,
        )

        return APiResponse(
            status_code=status_code,
            data=data,
            headers=response.headers,
            header_status_description=response.headers.get("x-status-description"),
            header_pagination=response.headers.get("x-pagination"),
            header_warning=response.headers.get("x-warning"),
        )

    def _post_process_response(
        self,
        response: Response,
        response_format: ResponseFormatEnum = ResponseFormatEnum.JSON,
        saved_path: Path = None,
    ):
        match response_format:
            case response_format.JSON:
                data = response.json()
                if saved_path is not None:
                    save_json_file(saved_path, data)
            case response_format.BYTES:
                data = response.content
                if saved_path is not None:
                    save_file(saved_path, data)
            case response_format.NO_CONTENT:
                data = None
            case _:
                raise ValueError(f"{response_format=} unknown")

        return response.status_code, data

    def assert_bytes_except_row(
        self,
        first: bytes,
        second: bytes,
        exception_row_prefix: bytes,
    ):
        first = first.replace(b"\r\n", b"\n")
        second = second.replace(b"\r\n", b"\n")
        first_rows = first.split(b"\n")
        second_rows = second.split(b"\n")
        assert len(first_rows) == len(second_rows), "Rows count mismatch"
        for first_row, second_row in zip(first_rows, second_rows):
            if first_row.startswith(exception_row_prefix) or second_row.startswith(
                exception_row_prefix
            ):
                continue  # Skip rows starting with exception_row_prefix
            self.assertEqual(first_row, second_row)

    def assertDictEqualExceptKeys(
        self,
        dict1: dict,
        dict2: dict,
        ignored_keys: list[str] = None,
    ):
        """Asserts that two dictionaries are equal except for the specified keys."""
        if ignored_keys is None:
            ignored_keys = []

        dict1_filtered = {k: v for k, v in dict1.items() if k not in ignored_keys}
        dict2_filtered = {k: v for k, v in dict2.items() if k not in ignored_keys}
        self.assertDictEqual(dict1_filtered, dict2_filtered)

    def assertNestedDictEqual(
        self,
        first: Dict,
        second: Dict,
        ignored_keys: List[str] = None,
    ):
        ignored_keys = [re.compile(item) for item in ignored_keys or []]

        deep_diff = DeepDiff(
            first,
            second,
            exclude_regex_paths=ignored_keys,
            ignore_numeric_type_changes=True,
        )

        assert not deep_diff, f"Dictionaries are not equal: {deep_diff}"

    def assertListOfDictEqual(
        self,
        first: list[dict],
        second: list[dict],
        ignored_keys: List[str] = None,
    ):
        self.assertEqual(len(first), len(second))
        for dict1, dict2 in zip(first, second):
            self.assertNestedDictEqual(dict1, dict2, ignored_keys)

    def assertResponseEqual(
        self,
        expected_response: ExpectedResponse,
        response: APiResponse,
        ignore_keys: bool = True,
    ):
        self.assertEqual(expected_response.status_code, response.status_code)

        self.assertEqual(
            expected_response.pagination, response.headers.get("x-pagination")
        )

        header_status_description = response.headers.get("x-status-description", [])
        if isinstance(header_status_description, str):
            header_status_description = json.loads(
                response.headers.get("x-status-description")
            )

            if not isinstance(header_status_description, list):
                header_status_description = [header_status_description]

        self.assertEqual(
            expected_response.status_description,
            header_status_description,
        )

        expected_response_warning = None
        if expected_response.warning is not None:
            expected_response_warning = "1" if expected_response.warning else "0"

        self.assertEqual(expected_response_warning, response.headers.get("x-warning"))

        if isinstance(expected_response.data, list):
            self.assertListOfDictEqual(
                expected_response.data,
                response.data,
                ignored_keys=self.get_ignored_keys() if ignore_keys else None,
            )
        elif isinstance(expected_response.data, dict):
            self.assertNestedDictEqual(
                expected_response.data,
                response.data,
                ignored_keys=self.get_ignored_keys() if ignore_keys else None,
            )
        else:
            self.assertEqual(
                expected_response.data,
                response.data,
            )



# MODULES
import json
from typing import Dict, List, Optional, TypedDict, Union


class ExtHeaders(TypedDict):
    pagination: Optional[str]
    status_description: Optional[Union[str, List[str]]]
    warning: Optional[bool]


def extend_headers(
    headers: Dict[str, str] | None = None,
    ext_headers: ExtHeaders | None = None,
) -> Optional[Dict[str, str]]:
    if ext_headers is None:
        return headers

    tmp_headers = {}

    access_control_expose_headers = []

    def add_ext_header(name: str, value: str) -> None:
        tmp_headers[name] = value
        access_control_expose_headers.append(name)

    if (pagination := ext_headers.get("pagination")) is not None:
        add_ext_header("x-pagination", pagination)
    if (status_description := ext_headers.get("status_description")) is not None:
        add_ext_header("x-status-description", json.dumps(status_description))
    if (warning := ext_headers.get("warning")) is not None:
        add_ext_header("x-warning", "1" if warning else "0")

    if tmp_headers is not None:
        headers = headers or {}

        headers["access-control-expose-headers"] = ", ".join(
            [
                *headers.get("access-control-expose-headers", "").split(", "),
                *access_control_expose_headers,
            ]
        )

        headers.update(tmp_headers)

    return headers



# MODULES
from pathlib import Path
from typing import Dict, List, Optional, Type, TypeVar, Union
import warnings

# PYDANTIC
from pydantic import BaseModel, ConfigDict, Field, computed_field

# LIBS
from alphaz_next.libs.file_lib import open_json_file

# MODELS
from alphaz_next.models.config.alpha_config import (
    ReservedConfigItem,
    replace_reserved_config,
)


class AlphaDatabaseConfigSchema(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    ini: bool = False
    init_database_dir_json: Optional[str] = Field(default=None)
    connect_args: Optional[Dict] = Field(default=None)
    create_on_start: bool = False
    view_names: List[str] = Field(default_factory=lambda: [])

    @computed_field
    @property
    def connection_string(self) -> str:
        raise NotImplementedError()


class AlphaDatabaseOracleConfigSchema(AlphaDatabaseConfigSchema):
    host: str
    username: str
    password: str
    port: int
    service_name: str
    type: str

    @computed_field
    @property
    def connection_string(self) -> str:
        return (
            f"oracle+cx_oracle://{self.username}:{self.password}@"
            f"{self.host}:{self.port}/{self.service_name}"
        )


class AlphaDatabaseSqliteConfigSchema(AlphaDatabaseConfigSchema):
    path: str

    @computed_field
    @property
    def connection_string(self) -> str:
        Path(self.path).parent.mkdir(parents=True, exist_ok=True)
        return f"sqlite:///{self.path}"


_T = TypeVar("_T", bound=BaseModel)


def create_databases_config(
    model: Type[_T],
    databases_config_path: Path,
    reserved_config: ReservedConfigItem,
) -> Optional[_T]:
    data = open_json_file(path=databases_config_path)

    configs: Dict[
        str : Union[AlphaDatabaseOracleConfigSchema, AlphaDatabaseOracleConfigSchema]
    ] = {}
    for k, v in data.items():
        db_type = v.get("type")
        v = replace_reserved_config(
            v,
            reserved_config=reserved_config,
        )
        match db_type:
            case "oracle":
                configs[k] = AlphaDatabaseOracleConfigSchema.model_validate(v)
            case "sqlite":
                configs[k] = AlphaDatabaseSqliteConfigSchema.model_validate(v)
            case _:
                warnings.warn(f"database type {db_type} is not supported")

    return model.model_validate(configs)



# MODULES
import datetime
from enum import Enum
import os
import re
import sys
import inspect
import logging
import unidecode
from logging.handlers import TimedRotatingFileHandler
from pathlib import Path
from typing import List, Optional

DEFAULT_FORMAT = (
    "{$date} - {$level:7} - {$pid:5} - {$file:>15}.{$line:<4} - {$name:<14}: $message"
)
DEFAULT_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


class LevelEnum(Enum):
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class AlphaLogger:
    def __init__(
        self,
        name: str,
        directory: str,
        level: int = logging.INFO,
        stream_output: bool = True,
        when: str = "midnight",
        interval: int = 1,
        backup_count: int = 10,
        file_name: Optional[str] = None,
        logging_formatter: str = DEFAULT_FORMAT,
        date_formatter: str = DEFAULT_DATE_FORMAT,
    ):
        self._name = name
        self._logging_formatter = logging_formatter
        self._date_formatter = date_formatter
        self._date_str = None

        if file_name is None:
            file_name = name

        error_name = "errors"
        warning_name = "warnings"
        monitoring_name = "monitoring"

        directory_path = Path(directory)
        directory_path.mkdir(parents=True, exist_ok=True)

        formatter = logging.Formatter()

        logger_config = {
            "level": level,
            "directory_path": directory_path,
            "when": when,
            "interval": interval,
            "backup_count": backup_count,
            "formatter": formatter,
        }

        self._logger = self._create_logger(
            name=name,
            file_name=file_name,
            stream_output=stream_output,
            **logger_config,
        )
        self._error_logger = self._create_logger(
            name=error_name,
            file_name=error_name,
            **logger_config,
        )
        self._warning_logger = self._create_logger(
            name=warning_name,
            file_name=warning_name,
            **logger_config,
        )
        self._monitoring_logger = self._create_logger(
            name=monitoring_name,
            file_name=monitoring_name,
            **logger_config,
        )

    def info(
        self,
        message: str,
        exc_info: Exception = None,
        stack_level: int = 1,
        monitor: Optional[str] = None,
    ):
        self._log(
            level=LevelEnum.INFO,
            message=message,
            stack=inspect.stack(),
            exc_info=exc_info,
            stack_level=stack_level,
            monitor=monitor,
        )

    def warning(
        self,
        message: str,
        exc_info: Exception = None,
        stack_level: int = 1,
        monitor: Optional[str] = None,
    ):
        self._log(
            level=LevelEnum.WARNING,
            message=message,
            stack=inspect.stack(),
            exc_info=exc_info,
            stack_level=stack_level,
            monitor=monitor,
        )

    def error(
        self,
        message: str,
        exc_info: Exception = None,
        stack_level: int = 1,
        monitor: Optional[str] = None,
    ):
        self._log(
            level=LevelEnum.ERROR,
            message=message,
            stack=inspect.stack(),
            exc_info=exc_info,
            stack_level=stack_level,
            monitor=monitor,
        )

    def critical(
        self,
        message: str,
        exc_info: Exception = None,
        stack_level: int = 1,
        monitor: Optional[str] = None,
    ):
        self._log(
            level=LevelEnum.CRITICAL,
            message=message,
            stack=inspect.stack(),
            exc_info=exc_info,
            stack_level=stack_level,
            monitor=monitor,
        )

    def _log(
        self,
        level: LevelEnum,
        message: str,
        stack: List[inspect.FrameInfo],
        exc_info=None,
        stack_level: int = 1,
        monitor: Optional[str] = None,
    ):
        self._date_str = self._get_current_date()

        message_formatted = self._get_formatted_message(
            message=message,
            level=level,
            stack=stack,
            stack_level=stack_level,
        )

        match level:
            case LevelEnum.INFO:
                self._logger.info(
                    message_formatted,
                    exc_info=exc_info,
                    stacklevel=stack_level,
                )
            case LevelEnum.WARNING:
                self._logger.warning(
                    message_formatted,
                    exc_info=exc_info,
                    stacklevel=stack_level,
                )
                self._warning_logger.warning(
                    message_formatted,
                    exc_info=exc_info,
                    stacklevel=stack_level,
                )
            case LevelEnum.ERROR:
                self._logger.error(
                    message_formatted,
                    exc_info=exc_info,
                    stacklevel=stack_level,
                )
                self._error_logger.error(
                    message_formatted,
                    exc_info=exc_info,
                    stacklevel=stack_level,
                )
            case LevelEnum.CRITICAL:
                self._logger.critical(
                    message_formatted,
                    exc_info=exc_info,
                    stacklevel=stack_level,
                )
                self._error_logger.critical(
                    message_formatted,
                    exc_info=exc_info,
                    stacklevel=stack_level,
                )

        if monitor is not None:
            message_formatted = self._process_monitoring_message(
                message=message_formatted,
                monitor=monitor,
            )
            self._monitoring_logger.info(
                message_formatted,
                exc_info=exc_info,
                stacklevel=stack_level,
            )

    def _get_formatted_message(
        self,
        message: str,
        level: LevelEnum,
        stack: List[inspect.FrameInfo],
        stack_level: int = 1,
    ):
        message_formatted = self._logging_formatter

        parameters = re.findall("\{\$([a-zA-Z0-9]*):?[0-9<>]*\}", message_formatted)

        parameters_values = []

        if stack_level >= len(stack):
            stack_level = len(stack) - 1
        caller = inspect.getframeinfo(stack[stack_level][0])

        structure = "$%s"
        keys = {
            "date": self._date_str,
            "pid": os.getpid(),
            "level": level.value.upper(),
            "name": self._name,
            "path": caller.filename,
            "file": caller.filename.split(os.sep)[-1].replace(".py", ""),
            "line": caller.lineno,
        }

        for parameter_name in parameters:
            if parameter_name in keys:
                message_formatted = message_formatted.replace(
                    structure % parameter_name, ""
                )
                parameters_values.append(keys[parameter_name])

        message_formatted = message_formatted.format(*parameters_values).replace(
            structure % "message", str(message)
        )

        return message_formatted

    def _get_current_date(self):
        current_date = datetime.datetime.now()
        return current_date.strftime(self._date_formatter)

    @classmethod
    def _process_monitoring_message(cls, message: str, monitor: str) -> str:
        return unidecode.unidecode(message.replace(message, f"[{monitor}] ({message})"))

    def _create_logger(
        self,
        name: str,
        level: int,
        directory_path: Path,
        file_name: str,
        when: str,
        interval: int,
        backup_count: int,
        formatter: logging.Formatter,
        stream_output: bool = False,
    ):
        logger = logging.getLogger(name=name)

        if logger.hasHandlers():
            return logger

        logger.setLevel(level)

        if stream_output:
            # Add a stream handler to log messages to stdout
            stream_handler = logging.StreamHandler(stream=sys.stdout)
            stream_handler.setLevel(level)
            stream_handler.setFormatter(formatter)
            logger.addHandler(stream_handler)

        # Add a file handler to log messages to a file
        time_rotating_handler = TimedRotatingFileHandler(
            filename=directory_path / f"{file_name}.log",
            when=when,
            interval=interval,
            backupCount=backup_count,
        )
        time_rotating_handler.setLevel(level)
        time_rotating_handler.setFormatter(formatter)
        logger.addHandler(time_rotating_handler)

        return logger



# MODULES
from datetime import datetime
from typing import List, Optional

# PYDANTIC
from pydantic import BaseModel, Field


class UserShortSchema(BaseModel):
    username: str
    last_activity: datetime
    permissions: List[str] = Field(default_factory=lambda: [])


class UserSchema(UserShortSchema):
    id: int
    registered_date: datetime
    email: Optional[str] = Field(default=None)
    short_login: Optional[str] = Field(default=None)
    full_name: Optional[str] = Field(default=None)
    location: Optional[str] = Field(default=None)
    country: Optional[str] = Field(default=None)
    region: Optional[str] = Field(default=None)
    disabled: bool



# MODULES
from typing import Any, Dict

# FASTAPI
from fastapi.responses import UJSONResponse as _UJSONResponse

# STARLETTE
from starlette.background import BackgroundTask

# CORE
from alphaz_next.core._base import extend_headers, ExtHeaders


class UJSONResponse(_UJSONResponse):
    def __init__(
        self,
        content: Any,
        status_code: int = 200,
        headers: Dict[str, str] | None = None,
        ext_headers: ExtHeaders | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None,
    ) -> None:
        headers = extend_headers(
            headers=headers,
            ext_headers=ext_headers,
        )

        super().__init__(content, status_code, headers, media_type, background)



# MODULES
from datetime import datetime
from typing import List, Optional

# MODELS
from alphaz_next.models.auth.user import UserSchema, UserShortSchema

GET_USER_PATH = "alphaz_next.auth.auth.get_user"
GET_API_KEY_PATH = "alphaz_next.auth.auth.get_api_key"


def get_mocked_user(
    id: int = 1,
    username: str = "foo",
    email: Optional[str] = "foo@st.com",
    short_login: Optional[str] = "bar",
    full_name: Optional[str] = "zoo",
    location: Optional[str] = None,
    country: Optional[str] = None,
    region: Optional[str] = None,
    disabled: bool = False,
    registered_date: datetime = datetime.now(),
    last_activity: datetime = datetime.now(),
    permissions: List[str] = [],
):
    return UserSchema(
        id=id,
        username=username,
        email=email,
        short_login=short_login,
        full_name=full_name,
        location=location,
        country=country,
        region=region,
        disabled=disabled,
        registered_date=registered_date,
        last_activity=last_activity,
        permissions=permissions,
    )


def get_mocked_short_user(
    username: str = "foo",
    last_activity: datetime = datetime.now(),
    permissions: List[str] = [],
) -> UserShortSchema:
    return UserShortSchema(
        username=username,
        last_activity=last_activity,
        permissions=permissions,
    )



# MODULES
import os
import pytz
import re
from typing import Dict, Generator, List, Optional, TypedDict
from pathlib import Path
from datetime import datetime
from logging import Logger

# SQLALCHEMY
from sqlalchemy import create_engine, orm, Table, text
from sqlalchemy.orm import Session, DeclarativeMeta
from sqlalchemy.schema import sort_tables
from sqlalchemy.inspection import inspect

# CONTEXTLIB
from contextlib import contextmanager

# LIBS
from alphaz_next.libs.file_lib import open_json_file


class _DataBaseConfigTypedDict(TypedDict):
    connection_string: str
    ini: bool
    init_database_dir_json: Optional[str]
    create_on_start: bool
    view_names: List[str]
    pool_size: int
    connect_args: Optional[Dict]


class AlphaDatabase:
    def __init__(
        self,
        databases_config: _DataBaseConfigTypedDict,
        base: DeclarativeMeta,
        logger: Logger,
    ) -> None:
        self._database_config = databases_config
        self._engine = create_engine(
            self._database_config.get("connection_string"),
            echo=False,
            connect_args=self._database_config.get("connect_args") or {},
        )
        self._base = base
        self._logger = logger

        self._session_factory = orm.sessionmaker(
            autocommit=False,
            autoflush=False,
            bind=self._engine,
        )

        if self._database_config.get("create_on_start"):
            self.create_database(view_names=self._database_config.get("view_names"))

    @property
    def ini(self):
        return self._database_config.get("ini")

    @property
    def init_database_dir_json(self):
        return self._database_config.get("init_database_dir_json")

    def create_database(
        self, view_names: List[Table] = None
    ) -> None:  # TODO find a way to get view names automatically
        insp = inspect(self._engine)
        current_view_names = [item.lower() for item in insp.get_view_names()]

        with self.session_factory() as session:
            for view in view_names or []:
                if view.name.lower() in current_view_names:
                    session.execute(text(f"DROP VIEW {view}"))

        self._base.metadata.create_all(self._engine)

    @contextmanager
    def session_factory(self) -> Generator[Session, None, None]:
        session = self._session_factory()
        try:
            yield session
        except Exception as ex:
            self._logger.error("Session rollback because of exception", exc_info=ex)
            session.rollback()
            raise
        finally:
            session.close()

    def init_tables_from_json_files(
        self,
        directory: Path,
        table_names: list[str],
        timezone="CET",
    ):
        if not (init := self.ini):
            raise ValueError(
                f"Unable to init database tables because {init=} in config"
            )

        def _post_process(data: dict):
            for key, value in data.items():
                if isinstance(value, str) and re.match(
                    r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d{3})?(Z|[+-]\d{2}:?\d{2})?",
                    value,
                ):
                    if value.endswith("Z"):
                        utc_dt = datetime.fromisoformat(value[:-1])
                        local_tz = pytz.timezone(timezone)
                        local_dt = utc_dt.replace(tzinfo=pytz.utc).astimezone(local_tz)
                        data[key] = local_dt
                    else:
                        data[key] = datetime.fromisoformat(value)

            return data

        if directory is None:
            directory = self.init_database_dir_json

        table_names = table_names or set()
        tables = {
            k: v for k, v in self._base.metadata.tables.items() if k in table_names
        }

        ordered_tables: list[Table] = sort_tables(tables.values())

        with self.session_factory() as session:
            for table in ordered_tables:
                table_name = str(table.name).upper()
                path = directory / f"{table_name}.json"

                session.execute(table.delete())

                if not os.path.exists(path):
                    self._logger.warning(
                        f"Failed to initialize {table_name=} due to the absence of the file at [{path}]."
                    )
                else:
                    raw_data = open_json_file(path=path)

                    raw_data = [_post_process(data) for data in raw_data]

                    session.execute(table.insert().values(raw_data))

                    self._logger.info(
                        f"Successfully initialized {table_name=} from the file at {path}."
                    )

                session.commit()

        return ordered_tables



# MODULES
from typing import List, Optional, Sequence

# FASTAPI
from fastapi import APIRouter, FastAPI
from fastapi.openapi.utils import get_openapi, BaseRoute
from fastapi.middleware.cors import CORSMiddleware

# DEPENDENCY_INJECTOR
from dependency_injector import containers


# MODELS
from alphaz_next.models.config.alpha_config import AlphaConfigSchema

# ELASTICAPM
from elasticapm.contrib.starlette import make_apm_client, ElasticAPM


def _custom_openapi(config: AlphaConfigSchema, routes: List[BaseRoute]):
    title = config.project_name.upper()
    if config.environment.lower() != "prod":
        title = f"{title} [{config.environment.upper()}]"

    openapi_schema = get_openapi(
        title=title,
        version=config.version,
        description=config.api_config.openapi.description,
        contact={
            "name": config.api_config.openapi.contact.name,
            "email": config.api_config.openapi.contact.email,
        },
        routes=routes,
    )

    return openapi_schema


def create_app(
    config: AlphaConfigSchema,
    routers: List[APIRouter],
    container: Optional[containers.DeclarativeContainer] = None,
    allow_origins: Sequence[str] = (),
    allow_methods: Sequence[str] = ("GET",),
    allow_headers: Sequence[str] = (),
    allow_credentials: bool = False,
) -> FastAPI:
    # APP
    app = FastAPI(
        title=config.project_name.upper(),
        version=config.version,
        docs_url=None,
        redoc_url=None,
    )
    app.container = container

    app.add_middleware(
        CORSMiddleware,
        allow_origins=allow_origins,
        allow_credentials=allow_credentials,
        allow_methods=allow_methods,
        allow_headers=allow_headers,
    )

    if config.api_config.apm is not None and config.api_config.apm.active:
        apm = make_apm_client(
            {
                "SERVICE_NAME": config.project_name,
                "ENVIRONMENT": config.api_config.apm.environment,
                "SERVER_URL": config.api_config.apm.server_url,
                "SERVER_CERT": config.api_config.apm.ssl_ca_cert,
                "VERIFY_SERVER_CERT": config.api_config.apm.ssl_verify,
            }
        )

        app.add_middleware(ElasticAPM, client=apm)

    for router in routers:
        app.include_router(router)

    app.openapi_schema = _custom_openapi(config=config, routes=app.routes)

    return app
